import os
from functools import partial
from datetime import datetime, timedelta
from dateutil.parser import parse as parse_date
import numpy as np
from scipy import ndimage as ndi
from scipy import stats

from tobac_flow.flow import create_flow, combine_flow
from tobac_flow.dataloader import goes_dataloader
from tobac_flow.dataset import (
    add_label_coords,
    add_step_labels,
    add_label_coords,
    flag_edge_labels,
    flag_nan_adjacent_labels,
    link_step_labels,
    calculate_label_properties,
)
from tobac_flow.analysis import (
    get_label_stats,
    weighted_statistics_on_labels,
    find_object_lengths,
    remap_labels,
    mask_labels,
)
from tobac_flow.label import flow_link_overlap, make_step_labels
from tobac_flow.detection import get_curvature_filter, get_growth_rate, get_peak_filter
from tobac_flow.utils import (
    create_dataarray,
    add_dataarray_to_ds,
    labeled_comprehension,
    slice_labels,
    linearise_field,
)

import argparse

parser = argparse.ArgumentParser(
    description="""Detect and track DCCs in GOES-16 ABI data"""
)
parser.add_argument("date", help="Date on which to start process", type=str)
parser.add_argument("hours", help="Number of hours to process", type=float)
parser.add_argument("-sat", help="GOES satellite", default=16, type=int)
parser.add_argument("-x0", help="Initial subset x location", default=0, type=int)
parser.add_argument("-x1", help="End subset x location", default=2500, type=int)
parser.add_argument("-y0", help="Initial subset y location", default=0, type=int)
parser.add_argument("-y1", help="End subset y location", default=1500, type=int)
parser.add_argument(
    "-sd",
    help="Directory to save preprocess files",
    default="../data/dcc_detect",
    type=str,
)
parser.add_argument("-gd", help="GOES directory", default="../data/GOES16", type=str)
parser.add_argument(
    "--extend_path",
    help="Extend save directory using year/month/day subdirectories",
    default=True,
    type=bool,
)
parser.add_argument(
    "--save_bt",
    help="Save brightness temperature field to output file",
    action="store_true",
)
parser.add_argument(
    "--save_wvd",
    help="Save water vapour difference field to output file",
    action="store_true",
)
parser.add_argument(
    "--save_swd",
    help="Save split window difference field to output file",
    action="store_true",
)
parser.add_argument(
    "--save_label_props",
    help="Save statistics of label properties to output file",
    action="store_true",
)
parser.add_argument(
    "--save_field_props",
    help="Save statistics of field properties to output file",
    action="store_true",
)
parser.add_argument(
    "--save_spatial_props",
    help="Save statistics of label spatial properties to output file",
    action="store_true",
)


args = parser.parse_args()
start_date = parse_date(args.date, fuzzy=True)
end_date = start_date + timedelta(hours=args.hours)

satellite = int(args.sat)
x0 = int(args.x0)
x1 = int(args.x1)
y0 = int(args.y0)
y1 = int(args.y1)
t_offset = 3

save_dir = args.sd
if not os.path.isdir(save_dir):
    try:
        os.makedirs(save_dir)
    except (FileExistsError, OSError):
        pass

save_name = "detected_dccs_G%02d_S%s_E%s_X%04d_%04d_Y%04d_%04d.nc" % (
    satellite,
    start_date.strftime("%Y%m%d_%H0000"),
    end_date.strftime("%Y%m%d_%H0000"),
    x0,
    x1,
    y0,
    y1,
)

save_path = os.path.join(save_dir, save_name)

print("Saving output to:", save_path)

goes_data_path = args.gd
if not os.path.isdir(goes_data_path):
    try:
        os.makedirs(goes_data_path)
    except (FileExistsError, OSError):
        pass


def main() -> None:

    print(datetime.now(), "Loading ABI data", flush=True)
    print("Loading goes data from:", goes_data_path, flush=True)

    bt, wvd, swd, dataset = goes_dataloader(
        start_date,
        end_date,
        n_pad_files=t_offset + 1,
        x0=x0,
        x1=x1,
        y0=y0,
        y1=y1,
        return_new_ds=True,
        satellite=satellite,
        product="MCMIP",
        view="C",
        mode=[3, 4, 6],
        save_dir=goes_data_path,
        replicate_path=True,
        check_download=True,
        n_attempts=1,
        download_missing=True,
    )

    print(datetime.now(), "Calculating flow field", flush=True)

    flow = combine_flow(
        create_flow(bt, model="Farneback", vr_steps=1, smoothing_passes=1),
        create_flow(wvd, model="Farneback", vr_steps=1, smoothing_passes=1),
        create_flow(swd, model="Farneback", vr_steps=1, smoothing_passes=1),
    )

    print(datetime.now(), "Detecting growth markers", flush=True)
    wvd_threshold = 0.25
    bt_threshold = 0.5
    overlap = 0.5
    subsegment_shrink = 0.0

    wvd_growth = get_growth_rate(flow, wvd, method="cubic")
    bt_growth = get_growth_rate(flow, -bt, method="cubic")

    wvd_curvature_filter = get_curvature_filter(wvd, direction="negative")
    bt_curvature_filter = get_curvature_filter(bt, direction="positive")

    wvd_peak_filter = get_peak_filter(wvd, sigma=0.5, direction="negative")
    bt_peak_filter = get_peak_filter(wvd, sigma=0.5, direction="positive")

    t_struct = np.zeros([3, 3, 3], dtype=bool)
    t_struct[:, 1, 1] = True
    bt_filter = flow.convolve(
        np.logical_or(bt_curvature_filter, bt_peak_filter).astype(int),
        structure=t_struct,
        method="nearest",
        fill_value=False,
        dtype=np.int32,
        func=partial(np.any, axis=0),
    )
    wvd_filter = flow.convolve(
        np.logical_or(wvd_curvature_filter, wvd_peak_filter).astype(int),
        structure=t_struct,
        method="nearest",
        fill_value=False,
        dtype=np.int32,
        func=partial(np.any, axis=0),
    )

    swd_filter = 1 - linearise_field(swd.data, 2.5, 7.5)

    wvd_markers = np.logical_and((wvd_growth * swd_filter) > wvd_threshold, wvd_filter)
    bt_markers = np.logical_and((bt_growth * swd_filter) > bt_threshold, bt_filter)

    s_struct = ndi.generate_binary_structure(3, 1)
    s_struct *= np.array([0, 1, 0])[:, np.newaxis, np.newaxis].astype(bool)

    combined_markers = ndi.binary_opening(
        np.logical_or.reduce([wvd_markers, bt_markers]), structure=s_struct
    )

    print("WVD growth above threshold: area =", np.sum(wvd_markers))
    print("BT growth above threshold: area =", np.sum(bt_markers))
    print("Detected markers: area =", np.sum(combined_markers))

    core_labels = flow.label(
        combined_markers, overlap=overlap, subsegment_shrink=subsegment_shrink
    )

    print("Initial core count:", np.max(core_labels))

    # Filter labels by length and wvd growth threshold
    core_label_lengths = find_object_lengths(core_labels)

    print(
        "Core labels meeting length threshold:", np.sum(core_label_lengths > t_offset)
    )

    core_label_wvd_mask = mask_labels(core_labels, wvd > -5)

    print("Core labels meeting WVD threshold:", np.sum(core_label_wvd_mask))

    combined_mask = np.logical_and(core_label_lengths > t_offset, core_label_wvd_mask)

    core_labels = remap_labels(core_labels, combined_mask)

    core_step_labels = slice_labels(core_labels)

    mode = lambda x: stats.mode(x, keepdims=False)[0]
    core_step_core_index = labeled_comprehension(
        core_labels, core_step_labels, mode, default=0
    )

    core_step_bt_mean = labeled_comprehension(
        bt, core_step_labels, np.nanmean, default=np.nan
    )

    core_step_t = labeled_comprehension(
        bt.t.data[:, np.newaxis, np.newaxis], core_step_labels, np.nanmin, default=0
    )

    def bt_diff_func(step_bt, pos):
        step_t = core_step_t[pos]
        args = np.argsort(step_t)

        step_bt = step_bt[args]
        step_t = step_t[args]

        step_bt_diff = (step_bt[:-t_offset] - step_bt[t_offset:]) / (
            (step_t[t_offset:] - step_t[:-t_offset])
            .astype("timedelta64[s]")
            .astype("int")
            / 60
        )

        return np.nanmax(step_bt_diff)

    core_bt_diff_mean = labeled_comprehension(
        core_step_bt_mean,
        core_step_core_index,
        bt_diff_func,
        default=0,
        pass_positions=True,
    )

    wh_valid_core = core_bt_diff_mean >= 0.5

    print("Core labels meeting cooling rate threshold:", np.sum(wh_valid_core))

    core_labels = remap_labels(core_labels, wh_valid_core)

    print("Final detected core count: n =", core_labels.max())

    # Now clean up used variables after core detection
    del (
        wvd_growth,
        bt_growth,
        wvd_curvature_filter,
        bt_curvature_filter,
        wvd_markers,
        bt_markers,
        combined_markers,
    )
    import gc

    gc.collect()

    print(datetime.now(), "Detecting thick anvil region", flush=True)
    # Detect anvil regions
    upper_threshold = -5
    lower_threshold = -15
    erode_distance = 2

    field = linearise_field(
        (wvd - (2 * np.maximum(swd - 5, 0))).data, lower_threshold, upper_threshold
    )

    structure = ndi.generate_binary_structure(3, 1)
    s_struct = structure * np.array([0, 1, 0])[:, np.newaxis, np.newaxis].astype(bool)

    # markers = np.logical_or(field>=upper_threshold, core_labels!=0)
    markers = field >= 1
    markers = ndi.binary_erosion(markers, structure=s_struct)
    field[markers] = 1
    mask = ndi.binary_erosion(
        field <= 0,
        structure=np.ones([3, 3, 3]),
        iterations=erode_distance,
        border_value=1,
    )

    edges = flow.sobel(field, direction="uphill", method="linear")
    marker_labels = flow.label(
        markers, overlap=overlap, subsegment_shrink=subsegment_shrink
    )

    print("Initial thick anvil markers: area =", np.sum(marker_labels != 0), flush=True)
    print("Initial thick anvil markers: n =", marker_labels.max(), flush=True)

    marker_label_lengths = find_object_lengths(marker_labels)
    marker_label_threshold = mask_labels(marker_labels, core_labels != 0)

    marker_labels = remap_labels(
        marker_labels,
        np.logical_and(marker_label_lengths > t_offset, marker_label_threshold),
    )

    print("Final thick anvil markers: area =", np.sum(marker_labels != 0), flush=True)
    print("Final thick anvil markers: n =", marker_labels.max(), flush=True)

    thick_anvil_labels = flow.watershed(
        edges - field, marker_labels, mask=mask, structure=structure
    )
    thick_anvil_labels *= ndi.binary_opening(
        thick_anvil_labels, structure=s_struct
    ).astype(int)

    print(datetime.now(), "Labelling thick anvil region", flush=True)
    thick_anvil_labels = flow_link_overlap(
        flow, make_step_labels(thick_anvil_labels), overlap=overlap
    )

    print(
        "Initial detected thick anvils: area =",
        np.sum(thick_anvil_labels != 0),
        flush=True,
    )
    print("Initial detected thick anvils: n =", thick_anvil_labels.max(), flush=True)

    marker_label_lengths = find_object_lengths(thick_anvil_labels)
    marker_label_threshold = mask_labels(thick_anvil_labels, core_labels != 0)

    thick_anvil_labels = remap_labels(
        thick_anvil_labels,
        np.logical_and(marker_label_lengths > t_offset, marker_label_threshold),
    )

    print(
        "Final detected thick anvils: area =",
        np.sum(thick_anvil_labels != 0),
        flush=True,
    )
    print("Final detected thick anvils: n =", thick_anvil_labels.max(), flush=True)

    print(datetime.now(), "Detecting thin anvil region", flush=True)
    # Detect thin anvil regions
    upper_threshold = -5
    lower_threshold = -10
    erode_distance = 2

    field = linearise_field(
        (wvd + (2 * np.maximum(swd - 5, 0))).data, lower_threshold, upper_threshold
    )

    structure = ndi.generate_binary_structure(3, 1)
    s_struct = structure * np.array([0, 1, 0])[:, np.newaxis, np.newaxis].astype(bool)

    # markers = np.logical_or(field>=upper_threshold, core_labels!=0)
    markers = thick_anvil_labels
    markers *= ndi.binary_erosion(markers != 0, structure=s_struct).astype(int)
    field[markers != 0] = 1

    mask = ndi.binary_erosion(
        field <= 0,
        structure=np.ones([3, 3, 3]),
        iterations=erode_distance,
        border_value=1,
    )

    edges = flow.sobel(field, direction="uphill", method="linear")

    thin_anvil_labels = flow.watershed(
        edges - field, markers, mask=mask, structure=structure
    )

    thin_anvil_labels *= ndi.binary_opening(
        thin_anvil_labels, structure=s_struct
    ).astype(int)

    print("Detected thin anvils: area =", np.sum(thin_anvil_labels != 0), flush=True)
    print("Detected thin anvils: n =", np.max(thin_anvil_labels), flush=True)

    print(datetime.now(), "Preparing output", flush=True)

    # Create output dataset
    # Core labels
    add_dataarray_to_ds(
        create_dataarray(
            core_labels,
            ("t", "y", "x"),
            "core_label",
            coords={"t": bt.t},
            long_name="labels for detected cores",
            units="",
            dtype=np.int32,
        ).sel(t=dataset.t),
        dataset,
    )

    # Thick anvil
    add_dataarray_to_ds(
        create_dataarray(
            thick_anvil_labels,
            ("t", "y", "x"),
            "thick_anvil_label",
            coords={"t": bt.t},
            long_name="labels for detected thick anvil regions",
            units="",
            dtype=np.int32,
        ).sel(t=dataset.t),
        dataset,
    )

    # Thin anvil
    add_dataarray_to_ds(
        create_dataarray(
            thin_anvil_labels,
            ("t", "y", "x"),
            "thin_anvil_label",
            coords={"t": bt.t},
            long_name="labels for detected thin anvil regions",
            units="",
            dtype=np.int32,
        ).sel(t=dataset.t),
        dataset,
    )

    # bt, wvd, swd = bt.sel(t=dataset.t), wvd.sel(t=dataset.t), swd.sel(t=dataset.t)

    add_step_labels(dataset)

    dataset = add_label_coords(dataset)

    link_step_labels(dataset)

    # Add data quality flags
    flag_edge_labels(dataset, start_date, end_date)
    flag_nan_adjacent_labels(dataset, bt.sel(t=dataset.t))

    core_max_bt_diff = core_bt_diff_mean[wh_valid_core]

    add_dataarray_to_ds(
        create_dataarray(
            core_max_bt_diff,
            ("core"),
            "core_max_BT_diff",
            long_name="Maximum change in core brightness temperature per minute",
            units="K/minute",
            dtype=np.float32,
        ),
        dataset,
    )

    if args.save_label_props:
        calculate_label_properties(dataset)

    if args.save_spatial_props:
        get_label_stats(dataset.core_label, dataset)
        get_label_stats(dataset.thick_anvil_label, dataset)
        get_label_stats(dataset.thin_anvil_label, dataset)

    if args.save_bt:
        add_dataarray_to_ds(
            bt.sel(t=dataset.t),
            dataset,
        )
    if args.save_wvd:
        add_dataarray_to_ds(
            wvd.sel(t=dataset.t),
            dataset,
        )
    if args.save_swd:
        add_dataarray_to_ds(
            swd.sel(t=dataset.t),
            dataset,
        )

    if args.save_field_props:
        # Add BT, WVD, SWD stats
        weights = np.repeat(dataset.area.data[np.newaxis, ...], dataset.t.size, 0)

        for field in (bt.sel(t=dataset.t), wvd.sel(t=dataset.t), swd.sel(t=dataset.t)):
            [
                add_dataarray_to_ds(da, dataset)
                for da in weighted_statistics_on_labels(
                    dataset.core_label,
                    field.compute(),
                    weights,
                    name="core",
                    dim="core",
                    dtype=np.float32,
                )
            ]

            [
                add_dataarray_to_ds(da, dataset)
                for da in weighted_statistics_on_labels(
                    dataset.thick_anvil_label,
                    field.compute(),
                    weights,
                    name="thick_anvil",
                    dim="anvil",
                    dtype=np.float32,
                )
            ]

            [
                add_dataarray_to_ds(da, dataset)
                for da in weighted_statistics_on_labels(
                    dataset.thin_anvil_label,
                    field.compute(),
                    weights,
                    name="thin_anvil",
                    dim="anvil",
                    dtype=np.float32,
                )
            ]

            [
                add_dataarray_to_ds(da, dataset)
                for da in weighted_statistics_on_labels(
                    dataset.core_step_label,
                    field.compute(),
                    weights,
                    name="core_step",
                    dim="core_step",
                    dtype=np.float32,
                )
            ]

            [
                add_dataarray_to_ds(da, dataset)
                for da in weighted_statistics_on_labels(
                    dataset.thick_anvil_step_label,
                    field.compute(),
                    weights,
                    name="thick_anvil_step",
                    dim="thick_anvil_step",
                    dtype=np.float32,
                )
            ]

            [
                add_dataarray_to_ds(da, dataset)
                for da in weighted_statistics_on_labels(
                    dataset.thin_anvil_step_label,
                    field.compute(),
                    weights,
                    name="thin_anvil_step",
                    dim="thin_anvil_step",
                    dtype=np.float32,
                )
            ]

    print(datetime.now(), "Saving to %s" % (save_path), flush=True)
    # Add compression encoding
    comp = dict(zlib=True, complevel=5, shuffle=True)
    for var in dataset.data_vars:
        dataset[var].encoding.update(comp)

    dataset.to_netcdf(save_path)

    dataset.close()
    bt.close()
    wvd.close()
    swd.close()


if __name__ == "__main__":
    start_time = datetime.now()
    print(start_time, "Commencing DCC detection", flush=True)

    print("Start date:", start_date)
    print("End date:", end_date)
    print("x0,x1,y0,y1:", x0, x1, y0, y1)
    print("Output save path:", save_path)
    print("GOES data path:", goes_data_path)

    main()

    print(
        datetime.now(),
        "Finished successfully, time elapsed:",
        datetime.now() - start_time,
        flush=True,
    )
