import numpy as np
import scipy.ndimage as ndi
from typing import Callable


def labeled_comprehension(
    field: np.ndarray,
    labels: np.ndarray[int],
    func: Callable,
    index: list[int] | None = None,
    dtype: type | None = None,
    default: float | None = None,
    pass_positions: bool = False,
) -> np.ndarray:
    """
    Wrapper for the scipy.ndimage.labeled_comprehension function

    Parameters
    ----------
    field : numpy.ndarray
        Data to apply the function overlap
    labels : numpy.ndarray
        The array of labeled regions
    func : function
        The function to apply to each labelled region
    index : list, optional (default : None)
        The label values to apply the comprehension over. Default value of None
            will apply the comprehension to all non-zero labels
    dtype : type, optional (default : None)
        The dtype of the output. Defaults to that of the input field
    default : scalar, optional (default : None)
        The value to return if a label does not exist
    pass_positions : bool, optional (default : False)
        If true, will pass the indexes of the label locations to the funtion in
            labeled_comprehension

    Returns
    -------
    comp : numpy.ndarray
        An array of the result of func applied to each labelled region included
            in index
    """
    if not dtype:
        dtype = field.dtype

    if index is None:
        index = np.unique(labels[labels != 0])

    comp = ndi.labeled_comprehension(
        field, labels, index, func, dtype, default, pass_positions
    )

    return comp


def apply_func_to_labels(
    labels: np.ndarray[int],
    *fields: tuple[np.ndarray],
    func: Callable = np.mean,
    index: None | list[int] = None,
    default: None | float = None,
) -> np.ndarray:
    """
    Apply a given function to the regions of an array given by an array of
        labels. Functions similar to ndi.labeled_comprehension, but may be more
        adaptable in some circumstances

    Parameters
    ----------
    labels : np.ndarray[int]
        labels of each region to apply function to
    *fields : tuple[np.ndarray]
        fields to give as arguments to each function call. Must have the same
            shape as labels
    func: Callable, optional (default: np.mean)
        function to apply over each region
    index: None | list[int], optional (default: None)
        list of indexes of regions in labels to apply function to. If None, will
            default to all integers between 1 and the maximum value in labels
    default: None | float, optional (default: None)
        default value to return in a region has no values
    """
    for field in fields:
        if labels.shape != field.shape:
            raise ValueError("Input labels and field do not have the same shape")
    if index is None:
        index = range(1, int(np.nanmax(labels) + 1))
    else:
        if np.max(index) > np.max(labels):
            raise ValueError("Index contains values that are not in labels!")
    bins = np.cumsum(np.bincount(labels.ravel()))
    args = np.argsort(labels.ravel())
    return np.array(
        [
            func(*[field.ravel()[args[bins[i - 1] : bins[i]]] for field in fields])
            if bins[i] > bins[i - 1]
            else default
            for i in index
        ]
    )


def flat_label(
    mask: np.ndarray[bool],
    structure: np.ndarray[int] = ndi.generate_binary_structure(3, 1),
    dtype=np.int32,
) -> np.ndarray[int]:
    """
    For a 3d+ field, return connected component labels that do not connect
        across the highest order dimension. In general this is used for finding
        labels at individual time steps.

    Parameters
    ----------
    mask : numpy.ndarray
        The (boolean) array of masked values to label into separate regions
    structure : numpy.ndarray, optional
        The structuring element to connect labels. This defaults to 1
            connectivity using ndi.generate_binary_structure(3,1).
    dtype : type, optional (default : np.int32)
        The dtype of the ouput labels

    Returns
    -------
    Output : numpy.ndarray
        The array of connected labelled regions

    See Also
    --------
    scipy.ndimage.label : this is the labelling routine used, with the
        connectivity of the highest dimension set to 0
    get_step_labels_for_label : find which step labels correspond to each input
        label
    """
    label_struct = structure.copy()
    label_struct[0] = 0
    label_struct[-1] = 0

    output = ndi.label(mask, structure=label_struct, output=dtype)[0]
    return output


def get_step_labels_for_label(
    labels: np.ndarray[int], step_labels: np.ndarray[int]
) -> list[int]:
    """
    Given the output from flat_label, and the original label array, find which
        step labels correspond to each original label

    Parameters
    ----------
    labels : numpy.ndarray
        The orginal array of labels
    step_labels : numpy.ndarray
        The array of labels split into separate steps

    Returns
    -------
    Output : numpy.ndarray
        The array of labelled regions at each time step

    See Also
    --------
    scipy.ndimage.label : this is the labelling routine used, with the
        connectivity of the highest dimension set to 0
    """
    bins = np.cumsum(np.bincount(labels.ravel()))
    args = np.argsort(labels.ravel())
    return [
        np.unique(step_labels.ravel()[args[bins[i] : bins[i + 1]]])
        if bins[i + 1] > bins[i]
        else None
        for i in range(bins.size - 1)
    ]


def relabel_objects(labels: np.ndarray[int]) -> np.ndarray[int]:
    """
    Given an array of labelled regions, renumber the labels so that they are
        contiguous integers

    Parameters
    ----------
    labels : numpy.ndarray
        The array of labels to renumber

    Returns
    -------
    new_labels : numpy.ndarray
        The regions labelled with contiguous integers
    """
    bins = np.cumsum(np.bincount(labels.ravel()))
    args = np.argsort(labels.ravel())
    new_labels = np.zeros_like(labels)
    counter = 1
    for i in range(bins.size - 1):
        if bins[i + 1] > bins[i]:
            new_labels.ravel()[args[bins[i] : bins[i + 1]]] = counter
            counter += 1
    return new_labels


def slice_labels(labels: np.ndarray[int]) -> np.ndarray[int]:
    """
    Given an array of labelled regions, will split these regions into separate
        labels along the leading dimension. In general this is used for finding
        the section of each label at each individual time step. Note that unlike
        flat_label the regions for each label at each time step will remain as a
        single label, even if it is not all connected at that step.

    Parameters
    ----------
    labels : numpy.ndarray
        The array of labels to split into individual time steps

    Returns
    -------
    step_labels : numpy.ndarray
        An array of labels corresponding the regions associated with the input
            labels at individual time steps
    """
    step_labels = np.zeros_like(labels)
    max_label = 0
    for i in range(labels.shape[0]):
        step_labels[i] = relabel_objects(labels[i])
        step_labels[i][np.nonzero(step_labels[i])] += max_label
        max_label = step_labels.max()
    return step_labels
